/*
 *  Copyright (C) 2000-2001 Marco Pesenti Gritti, Jorn Baayen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "galeon.h"
#include "galeon_types.h"
#include "favicon.h"
#include "mozilla.h"
#include "bookmarks.h"
#include "bookmarks_editor.h"
#include "favicon.h"
#include "eel-gconf-extensions.h"
#include "misc.h"

#include <string.h>
#include <gtk/gtkpixmap.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <libgnome/gnome-util.h>
#include <libgnomeui/gnome-window-icon.h>
#include <libgnomevfs/gnome-vfs.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

/** The favicon pixmap data cache */
static GHashTable *favicons_cache = NULL;

/** keyfile data */
static FILE *key_file = NULL;
static FILE *ie_file = NULL;
static GHashTable *key_hash = NULL;
static GHashTable *ie_hash = NULL;

/** local functions */
static void set_favicon (const gchar *site, const gchar *icon);
static void set_ie_favicon (const gchar *site, const gchar *icon);
static gchar *favicon_from_url (const gchar *site);
static gchar *strip_slash (const gchar *url);

/**
 * favicon_fetch: Parse an http URL ans get its favicon if there is one
 */
void
favicon_download (GaleonEmbed *embed, BookmarkItem *b, FaviconFetchFlags flags)
{
	GnomeVFSURI *vfs_uri;
	gchar *favicon_url = NULL, *favicon_path, *url;
	FaviconInfo *info;
	gboolean success = FALSE;
	gpointer key = NULL;
	PixmapData *val = NULL;
	gboolean is_ie = FALSE;

	if (b == NULL)
		url = embed->site_location;
	else
		url = b->url;

	if (url == NULL) return;
	
	vfs_uri = gnome_vfs_uri_new (url);
	if (vfs_uri == NULL) return;
	
	/* we have an embed, so we get the <link rel="shortcut icon"> favicon
	 * location */
	if (embed && (flags & FAVICON_FETCH_GET_LINK_ICON))
	{
		success = mozilla_get_favicon_location (embed, &favicon_url);
	}

	/* get host/favicon.ico if requested */
	if (!success && !favicon_url && (flags & FAVICON_FETCH_GET_BROKEN_ICON))
	{
		GnomeVFSURI *vfs_favicon_uri;
		
		while (gnome_vfs_uri_has_parent (vfs_uri))
		{
			GnomeVFSURI *vfs_parent_uri;
			vfs_parent_uri = gnome_vfs_uri_get_parent (vfs_uri);
			gnome_vfs_uri_unref (vfs_uri);
			vfs_uri = vfs_parent_uri;
		} 
		vfs_favicon_uri = gnome_vfs_uri_append_file_name (vfs_uri, 
								 "favicon.ico");
		favicon_url = gnome_vfs_uri_to_string (vfs_favicon_uri, 0);
		gnome_vfs_uri_unref (vfs_favicon_uri);

		is_ie = TRUE;
	}
	
	gnome_vfs_uri_unref (vfs_uri);

	if (favicon_url == NULL) return;

	favicon_path = favicon_filename (favicon_url);
	if (favicon_path == NULL)
	{
		g_free (favicon_url);
		return;
	}

	/* does it exist already ? */
	if (!(flags & FAVICON_FETCH_OVERWRITE) &&
	    g_file_exists (favicon_path))
	{
		/* put in the url - favicon file */
		if (!is_ie)
			set_favicon (url, favicon_path);
		else
			set_ie_favicon (url, favicon_path);

		/* update the drag handke pixmap in the toolbar */
		if (embed && embed->parent_window && !b)
		{
			favicon_update_drag_handle (embed, FALSE);
		}

		/* cleanup */
		g_free (favicon_url);
		g_free (favicon_path);

		return;
	}

	/* build the download information structure */
	info = g_new0 (FaviconInfo, 1);
	info->b = b;
	info->favicon_path = favicon_path;
	info->embed = embed;
	info->url = g_strdup (url);
	info->is_ie = is_ie;

	/* remove from cache */
	if (favicons_cache != NULL)
	{
		gchar *tmp = strip_slash (url);
		g_hash_table_lookup_extended (favicons_cache, tmp,
					      &key, (gpointer) &val);
		g_hash_table_remove (favicons_cache, tmp);
		g_free (tmp);
	}
	if (val)
	{
		if (val->pixmap) gdk_pixmap_unref (val->pixmap);
		if (val->mask) gdk_bitmap_unref (val->mask);
		g_free (val);
	}
	if (key) g_free (key);
	
	/* commence the download */
	mozilla_save_url (embed, favicon_url, favicon_path,
			  (b != NULL ? ACTION_FAVICON_EDITOR : ACTION_FAVICON),
			  (gpointer) info);

	g_free (favicon_url);
}

/**
 * favicon_download_completed: called when the favicon download finishes,
 * does all the required actions to get the favicon into the UI
 */
void 
favicon_download_completed (FaviconInfo *info)
{
	GdkPixbuf *test;
	BookmarkItem *b = info->b;
	gchar *favicon_path;

	g_assert (info != NULL);

	favicon_path = info->favicon_path;

	g_assert (favicon_path != NULL);

	/* check whether the download succeeded */
	if (!g_file_exists (favicon_path))
	{
		g_free (favicon_path);
		g_free (info->url);
		g_free (info);
		return;
	}

	/* check whether it's a legal image */
	test = gdk_pixbuf_new_from_file (favicon_path);
	if (test == NULL)
	{
		gnome_vfs_unlink (favicon_path);
		g_free (favicon_path);
		g_free (info->url);
		g_free (info);
		return;
	}
	else
	{
		gdk_pixbuf_unref (test);
	}

	/* put in the url - favicon file */
	if (!info->is_ie)
		set_favicon (info->url, info->favicon_path);
	else
		set_ie_favicon (info->url, info->favicon_path);

	/* update the drag handke pixmap in the toolbar */
	if (info->embed && g_list_find (all_embeds, info->embed) &&
	    info->embed->parent_window && !b)
	{
		favicon_update_drag_handle (info->embed, FALSE);
	}
  
	/* update the bookmarks editor if we were called from there */
	if (b && bookmarks_bookmark_exists (bookmarks_root, b))
	{
		bookmarks_dirty = TRUE;
		bookmarks_toolbars_check_update (b);
		bookmarks_editor_update_tree_items (b);
	}

	g_free (favicon_path);
	g_free (info->url);
	g_free (info);
}

/**
 * favicon_filename: parse a URL and get the filename of the associated favicon
 */
gchar *
favicon_filename (gchar *favicon_url)
{
	GnomeVFSURI *vfs_uri;
	gchar *result, *slashpos, *filename, *tmp;

	vfs_uri = gnome_vfs_uri_new (favicon_url);
	if (vfs_uri == NULL)
	{
		g_free (favicon_url);
		return NULL;
	}
	
	filename = gnome_vfs_uri_to_string (vfs_uri, 
				    GNOME_VFS_URI_HIDE_USER_NAME |
				    GNOME_VFS_URI_HIDE_PASSWORD |
				    GNOME_VFS_URI_HIDE_TOPLEVEL_METHOD |
				    GNOME_VFS_URI_HIDE_FRAGMENT_IDENTIFIER);

	while ((slashpos = strstr (filename, "/")) != NULL) *slashpos = '_';
	if (filename[strlen (filename) - 1] == '_')
		tmp = g_strndup (filename, strlen (filename) - 1);
	else
		tmp = g_strdup (filename);
	g_free (filename);

	result = g_strconcat (g_get_home_dir (), "/.galeon/favicons/",
			      tmp, NULL);
	g_free (tmp);

	gnome_vfs_uri_unref (vfs_uri);

	return result;
}

/**
 * favicon_update_drag_handle: update the drag handle pixmap in the toolbar
 */
void
favicon_update_drag_handle (GaleonEmbed *embed, gboolean fetch)
{
	const PixmapData *drag_pixmap;

	drag_pixmap = favicon_get_pixmap (embed->site_location);

	/* update the drag handle pixmap */
	if (embed->is_active && embed->parent_window
	    && embed->parent_window->drag_pixmap != NULL &&
	    drag_pixmap->pixmap !=
	    	GTK_PIXMAP (embed->parent_window->drag_pixmap)->pixmap)
	{
		gtk_pixmap_set (GTK_PIXMAP
				(embed->parent_window->drag_pixmap), 
				drag_pixmap->pixmap, 
				drag_pixmap->mask);
	}

	/* update icon in notebook */
	if (embed && embed->notebook_icon)
	{
		/* handle all different configs */
		switch (eel_gconf_get_integer (CONF_TABS_TABBED_FAVICONS))
		{
		case 0:
			if (GTK_WIDGET_VISIBLE (embed->notebook_icon))
			{
				gtk_widget_hide (embed->notebook_icon);
				embed_update_tab_size (embed);
			}
			break;
		case 2:
			if (drag_pixmap == site_pixmap_data)
			{
				if (GTK_WIDGET_VISIBLE (embed->notebook_icon))
				{
					gtk_widget_hide (embed->notebook_icon);
					embed_update_tab_size (embed);
				}
				break;
			}
			/* fall through */
		default:
			if (!GTK_WIDGET_VISIBLE (embed->notebook_icon))
			{
				gtk_widget_show (embed->notebook_icon);
				embed_update_tab_size (embed);
			}
			if (drag_pixmap->pixmap ==
			    GTK_PIXMAP (embed->notebook_icon)->pixmap)
			{
				break;
			}
			gtk_pixmap_set (GTK_PIXMAP (embed->notebook_icon), 
        	        	        drag_pixmap->pixmap, drag_pixmap->mask);
			break;
		}
	}

	/* download the favicon if it's not there yet and we're asked
	 * to do so */
	if (drag_pixmap == site_pixmap_data && fetch)
	{
		if (eel_gconf_get_boolean (CONF_GENERAL_IE_FAVICONS))
		{
			favicon_download (embed, NULL, FAVICON_FETCH_GET_LINK_ICON |
						       FAVICON_FETCH_GET_BROKEN_ICON);
		}
		else
		{
			favicon_download (embed, NULL, FAVICON_FETCH_GET_LINK_ICON);
		}
	}
}

/**
 * favicon_get_pixmap: Parse an http URL ans return the appropriate icon 
 */
const PixmapData *
favicon_get_pixmap (const gchar *url)
{
	PixmapData *result = NULL;
	gchar *favicon_path;

	if (url == NULL || strcmp (url, "") == 0 ||
	    strcmp (url, "about:blank") == 0)
	{
		return site_pixmap_data;
	}

	/* get the favicon filename */
	favicon_path = favicon_from_url (url);
	if (!favicon_path) return site_pixmap_data;
	
	/* check if it's cached already, if so, return the cached pixmap */
	if (favicons_cache != NULL)
	{
		result = g_hash_table_lookup (favicons_cache, favicon_path);
		/* perhaps we need to add a reference to the GdkPixmap
		   and the GdkBitmap? */
		if (result != NULL) return result;
	}
	
	/* if the favicon exists, make the pixmap and cache it */
	if (favicon_path && g_file_exists (favicon_path))
	{
		GdkPixbuf *pixbuf;
		
		pixbuf = gdk_pixbuf_new_from_file (favicon_path);
		if (pixbuf != NULL)
		{
			if (gdk_pixbuf_get_width (pixbuf) > 16 || 
			    gdk_pixbuf_get_height (pixbuf) > 16)
			{
				GdkPixbuf *scaled_icon = 
					gdk_pixbuf_scale_simple (pixbuf, 
						16, 16, GDK_INTERP_NEAREST);
				gdk_pixbuf_unref (pixbuf);
				pixbuf = scaled_icon;
			}

			/* threshold out alpha if possible */
			if (gdk_pixbuf_get_has_alpha       (pixbuf) &&
			    gdk_pixbuf_get_bits_per_sample (pixbuf) == 8 &&
			    gdk_pixbuf_get_colorspace      (pixbuf) ==
			    	GDK_COLORSPACE_RGB &&
			    gdk_pixbuf_get_n_channels      (pixbuf) == 4)
			{
				pixbuf_threshold_alpha (pixbuf, FALSE);
			}
			
			result = g_new (PixmapData, 1);
			gdk_pixbuf_render_pixmap_and_mask (pixbuf,
							   &(result->pixmap),
							   &(result->mask),
							   100);
			gdk_pixbuf_unref (pixbuf);
	
			if (favicons_cache == NULL) 
			{
				favicons_cache = g_hash_table_new (g_str_hash,
							 	   g_str_equal);
			}

			g_hash_table_insert (favicons_cache, favicon_path, result);
		}
		else
		{
			result = site_pixmap_data;
		}
	}
	else
	{
		result = site_pixmap_data;
	}

	return result;
}

/**
 * favicon_key_files_init: initialise the favicon keyfiles
 */
void
favicon_key_files_init (void)
{
	gchar *line = NULL, *filename = NULL;
	
	/*** load the <link> favicon keyfile ***/
	
	/* init the key hashtable */
	key_hash = g_hash_table_new (g_str_hash, g_str_equal);

	/* open the keyfile */
	filename = g_strconcat (g_get_home_dir (),
				"/.galeon/favicons.keys", NULL);
	key_file = fopen (filename, "r");

	/* read file */
	while (key_file &&
	       (line = read_line_from_file (key_file)) && strlen (line) > 0)
	{
		gchar **list;

		/* initialise line */
		if (line[strlen(line) - 1] == '\n')
		{
			line[strlen(line) - 1] = ' ';
		}
		list = g_strsplit (line, " ", 0);

		/* sanity checks */
		if (list[0] != NULL && list[1] != NULL)
		{
			if (g_file_exists (list[1]))
			{
				/* put in into the hashtable */
				g_hash_table_insert (key_hash,
						     g_strdup (list[0]),
						     g_strdup (list[1]));
			}
		}

		/* cleanup */
		g_strfreev (list);
		g_free (line);
	}
	g_free (line);

	/* open for writing now */
	if (key_file) fclose (key_file);
	key_file = fopen (filename, "a");

	g_free (filename);

	/*** load the host/favicon.ico keyfile ***/
	
	/* init the key hashtable */
	ie_hash = g_hash_table_new (g_str_hash, g_str_equal);

	/* open the keyfile */
	filename = g_strconcat (g_get_home_dir (),
				"/.galeon/favicons.iekeys", NULL);
	ie_file = fopen (filename, "r");

	/* read file */
	while (ie_file &&
	       (line = read_line_from_file (ie_file)) && strlen (line) > 0)
	{
		gchar **list;

		/* initialise line */
		if (line[strlen(line) - 1] == '\n')
		{
			line[strlen(line) - 1] = ' ';
		}
		list = g_strsplit (line, " ", 0);

		/* sanity checks */
		if (list[0] != NULL && list[1] != NULL)
		{
			if (g_file_exists (list[1]))
			{
				/* put in into the hashtable */
				g_hash_table_insert (ie_hash,
						     g_strdup (list[0]),
						     g_strdup (list[1]));
			}
		}

		/* cleanup */
		g_strfreev (list);
		g_free (line);
	}
	g_free (line);

	/* open for writing now */
	if (ie_file) fclose (ie_file);
	ie_file = fopen (filename, "a");

	g_free (filename);
}

/**
 * favicon_key_files_close: close the keyfiles
 */
void
favicon_key_files_close (void)
{
	/* close the keyfiles */
	fclose (key_file);
	fclose (ie_file);
}

/**
 * set_favicon: cache favicon filename for a site
 */
static void
set_favicon (const gchar *site, const gchar *icon)
{
	gchar *tmp;
	
	if (!g_file_exists (icon)) return;
	
	tmp = strip_slash (site);

	/* check it's not already there */
	if (g_hash_table_lookup (key_hash, tmp) == NULL)
	{
		/* put it in the keyfile */
		gchar *key = g_strdup_printf ("%s %s\n", tmp, icon); 
		fputs (key, key_file);
		g_free (key);
		fflush (key_file);

		/* put it in the hashtable */
		g_hash_table_insert (key_hash, tmp, g_strdup (icon));
	}
	else
	{
		g_free (tmp);
	}
}

/**
 * set_ie_favicon: cache ie favicon filename for a site
 */
static void
set_ie_favicon (const gchar *site, const gchar *icon)
{
	GnomeVFSURI *vfs_uri;
	gchar *tmp;
	
	if (!g_file_exists (icon)) return;
	
	vfs_uri = gnome_vfs_uri_new (site);
	if (vfs_uri == NULL) return;
	tmp = g_strdup (gnome_vfs_uri_get_host_name (vfs_uri));
	gnome_vfs_uri_unref (vfs_uri);
	if (tmp == NULL) return;

	/* check it's not already there */
	if (g_hash_table_lookup (ie_hash, tmp) == NULL)
	{
		/* put it in the keyfile */
		gchar *key = g_strdup_printf ("%s %s\n", tmp, icon); 
		fputs (key, ie_file);
		g_free (key);
		fflush (ie_file);

		/* put it in the hashtable */
		g_hash_table_insert (ie_hash, tmp, g_strdup (icon));
	}
	else
	{
		g_free (tmp);
	}
}

/**
 * favicon_from_url: get favicon filename for a site
 */
static gchar *
favicon_from_url (const gchar *site)
{
	/* get a sane url */
	gchar *tmp = strip_slash (site);

	/* first look for a <link> favicon */
	gchar *ret = g_hash_table_lookup (key_hash, tmp);
	g_free (tmp);

	/* no <link> favicon, look for a host/favicon.ico */
	if (ret == NULL)
	{
		GnomeVFSURI *vfs_uri;
	
		vfs_uri = gnome_vfs_uri_new (site);
		if (vfs_uri == NULL) return NULL;

		tmp = (gchar *) gnome_vfs_uri_get_host_name (vfs_uri);

		if (tmp) ret = g_hash_table_lookup (ie_hash, tmp);

		gnome_vfs_uri_unref (vfs_uri);
	}

	/* we're done */
	return ret;
}

/**
 * strip_slash: strip the trailing slash and hash from a filename/url
 */
static gchar *
strip_slash (const gchar *url)
{
	gchar *ret, *ret2, *tmp;

	/* get rid of trailing slash */
	if (url[strlen (url) - 1] == '/')
		ret = g_strndup (url, strlen (url) - 1);
	else
		ret = g_strdup (url);

	/* get rid of #section */
	tmp = strstr (ret, "#");
	if (tmp)
	{
		ret2 = g_strndup (ret, strlen (ret) - strlen (tmp));
		g_free (ret);
	}
	else
	{
		ret2 = ret;
	}

	return ret2;
}
